=begin

Copyright 2011-2013, Andreas Eisenbarth
All Rights Reserved

Permission to use, copy, modify, and distribute this software for
any purpose and without fee is hereby granted, provided that the above
copyright notice appear in all copies.

THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

Description:  It imports KML Files (Keyhole Markup Language) containing
              geo-referenced data.
Name:         KMLExporter.rb
Author:       Andreas Eisenbarth
Usage:        (menu) Plugins → KML Tools → Export KML File
Description:  This plugin either exports the current selection or the complete
              model to KML Files (Keyhole Markup Language) for Google Earth.
              The following SketchUp objects can be exported:

              Groups, Components                 as  <Model> or <MultiGeometry>

              grouped horizontal Faces,
              grouped extruded horizontal Faces  as  <Polygon> or <Model>

              grouped Edges, Arcs, Circles       as  <LineString> or <gx:Track> or <Model>

              ConstructionPoints, leader Texts   as  <Point>

              Images, grouped Images,
              grouped horizontal quadfaces       as  <GroundOverlay> or <Model>

              Scenes (Pages), Animation          as  <Camera>, <LookAt>, <gx:Tour>

              The model must be georeferenced first.
              Groups should not contain other groups,
              only entities on visible layers can be exported as <Model>.
Version:      2.0.0-dev
Date:         31.05.2013

TODO:
• support Attributes, export layers as Folders
=end
require 'sketchup.rb'
# TODO: remove requirement of dynamic components if possible.
unless Sketchup::require 'su_dynamiccomponents.rb'
  Sketchup::require 'dynamiccomponents.rb'
end



module AE



module KMLTools



DIR = File.dirname(__FILE__) unless defined?(self::DIR)
require(File.join(DIR, "Translate.rb"))
TRANSLATE = Translate.new("KMLTools", File.join(DIR, "lang")) unless defined?(self::TRANSLATE)
require(File.join(DIR, "Dialog.rb"))
require(File.join(DIR, "Options.rb"))
@debug ||= true
attr_accessor :debug



class Exporter # < Sketchup::Exporter # Hey SketchUp 2014!

  EXPORT_SUCCESS                        = 0
  EXPORT_FAILED                         = 1
  EXPORT_CANCELLED                      = 2
  EXPORT_FILE_NOT_FOUND                 = 4
  EXPORT_SKETCHUP_VERSION_NOT_SUPPORTED = 5

  @@options ||= Options.new("KMLExporter", {
    :export_group             => true,
    :group_as                 => :model, # :model, :multi_geometry
    :model_condense_textures  => true,
    :model_exclude_gearth     => true,
    :export_curve             => true,
    :curve_as                 => :line_string, # :line_string, :track
    :export_hface_as_polygon  => true,
    :hface_extrude            => true,
    :export_himage_as_ground_overlay => true,
    :himage_exclude_gearth    => true,
    # :export_photomatch      => false, # Not yet implemented because of lacking API methods to get photomatch image.
    :export_cpoint            => true,
    :export_text              => true,
    :export_page              => true,
    :page_as                  => :camera, # :camera, :look_at
    :pages_as_tour            => true,
    :altitude_mode            => :relativeToGround # :relativeToGround, :absolute, clampToGround
  })

  def initialize(options={}) # TODO: since we are asynchronous, we need :on_success, :on_failure
    @options_dlg = nil
    @@options.update(options)
    @quiet = (options[:quiet] == true)
    @log = []
    # A list of features and how many of them were exported.
    @failed = {}
    @exported = {}
  end

  def description
    return "Keyhole Markup Language (*.kml)"
  end

  def file_extension
    return "kml"
  end

  def id
    return "com.sketchup.exporters.ae.kml"
  end

  def supports_options?
    return true
  end

  def do_options
    begin
      features = scan_for_features
      show_options(features)
    rescue Exception => e
      $stderr.write(e.message)
      $stderr.write(e.backtrace)
    end
  end

  def write_file(path, status=true)
    @model = Sketchup.active_model
    # Check for georeference.
    if (@model.respond_to?(:georeferenced?) &&
          !@model.georeferenced? ) ||
          @model.get_attribute("GeoReference", "Longitude").nil? # SU7.1+
      UI.messagebox(TRANSLATE["The model needs to be geo-referenced first."], MB_OK)
      return EXPORT_CANCELLED
    end
    # Scan the model for features that can be exported.
    features = scan_for_features unless @scanned_for_features
    # Validate the file path.
    return EXPORT_FILE_NOT_FOUND unless path.is_a?(String)
    file = File.expand_path(path)
    return EXPORT_FILE_NOT_FOUND unless File.directory?(File.dirname(path))
    # Write the file.
    main(path)
    return EXPORT_SUCCESS
  rescue Exception => e
    $stderr.write(e.message)
    $stderr.write(e.backtrace.join("\n"))
    # Restore the selection.
    @selection.each{ |s| @model.selection.add s }
    return EXPORT_FAILED
  end



  def reset
    @scanned_for_features = false
    @data_dir = nil
    @file = nil
  end



private



# Scan model for features that can be exported and sort them into categories.
#
def scan_for_features
  @scanned_for_features = false
  @model = Sketchup.active_model

  # Cache the selection so that it can later be restored, because we need to
  # clear it when exporting parts of the model as Collada.
  @selection = Set.new
  @model.selection.each{ |s| @selection.insert(s) }

  # Either export complete model or only selection.
  entities = (@selection.empty?) ? @model.entities : @selection

  # Sort entities into features.
  @text   = []
  @cpoint = []
  @group  = []
  @himage = []
  @hface  = []
  @curve  = []
  edges  = []
  # Collect an array of names of features that were found in the model.
  features = []

  # Cache the paths to exported resources so we don't get duplicates.
  @exportedResources = {}

  # Absolute elevation above sea level of origin, in [m]
  @elevation = 0

  entities.each{ |entity|

    case entity

    # Text
    when Sketchup::Text
      if entity.has_leader? # but no screen text
        @text.push(entity)
        features.push(text)
      end

    # Construction point
    when Sketchup::ConstructionPoint
      @cpoint.push(entity)
      features.push(:cpoint)

    # Curve or edges:
    # Since element is only edge, collect all edges and add later to @curve.
    when Sketchup::Edge
      edges.push(entity) # temporary array

    # Horizontal image
    when Sketchup::Image
      definition = Utilities::Image.get_definition(entity)
      if e.normal == [0,0,1] # This considers tolerance ✓
        @himage.push({:entity => entity, :name => definition.name, :description => definition.description});
        features.push(:himage)
      else
        @group.push({:entity => entity, :name => definition.name, :description => definition.description})
        features.push(:model)
      end

    # Component
    when Sketchup::ComponentInstance
      name = (!entity.name.empty?) ? entity.name : entity.definition.name
      @group.push({:entity => entity, :name => name, :description => entity.definition.description})
      features.push(:model)

    # Check groups for their entities.
    when Sketchup::Group
      # TODO: Check for Folder attribute and export as folder
      name = (!entity.name.empty?) ? entity.name : entity.entities.parent.name
      face = entity.entities.find{ |x| x.is_a?(Sketchup::Face) }
      edges = entity.entities.grep(Sketchup::Edge)
      faces = entity.entities.grep(Sketchup::Face)

      # Horizontal image
      if entity.entities[0].is_a?(Sketchup::Image) &&
         entity.entities.length == 1 && entity.entities[0].normal.samedirection?(Z_AXIS)
        @himage.push({:entity => entity.entities[0], :name => name, :description => entity.description})
        features.push(:himage)

      # Horizontal textured quadface
      elsif face && entity.entities.length == 5 && face.edges.length == 4 &&
        face.material != nil && face.normal.transform(entity.transformation.inverse).samedirection?(Z_AXIS)
        @himage.push({:entity => face, :name => name, :description => entity.description, :parent => entity})
        features.push(:himage)

      # Curve or edges
      elsif entity.entities.length == edges.length
        @curve.push({:entities => entity.entities.to_a, :name => name, :description => entity.description, :parent => entity})
        features.push(:curve)

      # Extruded curve
      elsif edges.length == (3 * faces.length + 1) &&
          faces.inject(true){ |is_vertical, f| is_vertical && f.normal.transform(entity.transformation.inverse).perpendicular?(Z_AXIS) }
        # Find the upper edges. First reject vertical edges. Since the extrusion
        # is vertical, the remaining edges form pairs that lay on top of each
        # other and we pick the one with higher z value.
        edges = edges.reject{ |edge| edge.line[1].parallel?(Z_AXIS) }.
          sort_by{ |edge| (edge.start.position + edge.end.position.to_a).to_a }
        upper_edges = (0...(edges.length/2)).map{ |i| edges[2*i+1] }
        @curve.push({:entities => upper_edges, :name => name, :description => entity.description, :parent => entity, :extrude => true })
        features.push(:curve)

      # Extruded polygon
      elsif edges.length == 3 * (faces.length - 2) &&
          faces.inject(true){ |is_vertical_or_horizontal, f| v = f.normal.transform(entity.transformation.inverse); is_vertical_or_horizontal && (v.parallel?(Z_AXIS) || v.perpendicular?(Z_AXIS) ) }
        # Find the upper face(s).
        faces = faces.find_all{ |face| face.normal.transform(entity.transformation.inverse).parallel?(Z_AXIS) }
        upper_faces = faces.reject{ |face| c = face.bounds.center; !!(faces-[face]).find{ |f| c2 = f.bounds.center; c2.x == c.x && c2.y == c.y && c2.z > c.z } }
        @hface.push({:entities => upper_faces, :name => name, :description => entity.description, :parent => entity, :extrude => true })
        features.push(:hface)

      # Polygon (flat)
      elsif edges.length == faces.map{ |f| f.edges }.flatten.uniq.length &&
          faces.inject(true){ |is_horizontal, f| is_horizontal && f.normal.transform(entity.transformation.inverse).parallel?(Z_AXIS) }
        @hface.push({:entities => faces, :name => name, :description => entity.description, :parent => entity})
        features.push(:hface)

      # Normal group
      else
        @group.push({:entity => entity, :name => entity.name, :description => entity.description})
        features.push(:model)

      end # if e.entities
    end # case
  } # each

  # Page
  if @model.pages.count > 0
    features.push(:page)
  end

  features.uniq!
  features.push(:altitudeMode) unless features.empty?

  @scanned_for_features = true
  return features
end # def scan_for_features



# Ask the user which categories to export (and other options).
# @param [Array<String>] features  a list of features/categories which the dialog will offer to select.
#
def show_options(features=[])
  options_dlg_title = (@selection && @selection.empty?) ? TRANSLATE["Export to KML"] : TRANSLATE["Export selection to KML"]
  @options_dlg = AE::KMLTools::Dialog.new(options_dlg_title, false, "AE_KMLExporter", 550, 400, 300, 300, true)
  html_path = File.join(DIR, "html", "KMLExporterOptions.html")
  @options_dlg.set_file(html_path)

  # Callbacks.

  @options_dlg.on_show{ |dlg, param|
    TRANSLATE.webdialog(dlg)
    dlg.call_function("AE.KMLExporter.initialize", @@options.get_all, features)

    # Get the absolute elevation either from the model or a web service.
    @elevation = 0
    alt = @model.get_attribute("GeoReference", "ModelTranslationZ")
    if alt.is_a?(Numeric)
      @elevation = -alt.to_m
    else
      lat = @model.get_attribute("GeoReference","Latitude")
      lon = @model.get_attribute("GeoReference", "Longitude")
      dlg.call_function("AE.KMLExporter.getElevation", lat, lon)
    end
  }

  @options_dlg.on("update_options") { |dlg, options|
    @@options.update(options)
  }

  @options_dlg.on("receive_elevation") { |dlg, options|
    if elevation.is_a?(Numeric) && elevation != 9999
      @elevation = elevation
    end
  }

  @options_dlg.show_modal
  return @options_dlg
end # def options_dialog



# Evaluate the options, call each feature's export method and write the file.
# @param [String] path  a path to an existing valid kml file
#
def main(path)
  path = File.expand_path(path.unpack("U*").map{ |c| c.chr }.join)
  filename = File.basename(path, File.extname(path))
  # Directory for files associated with the KML.
  @data_dir = File.join( File.dirname(path), filename )
  Dir.mkdir(@data_dir) unless File.exist?(@data_dir) # TODO: do this on demand
  XMLFile.open(path, "w"){ |file|
    @file = file # TODO
    file.puts '<?xml version="1.0" encoding="UTF-8"?>'
    file.indent('kml xmlns="http://www.opengis.net/kml/2.2" xmlns:gx="http://www.google.com/kml/ext/2.2"'){
      file.indent(:Document){
        file.set(:name, filename)
        # Don't expand the whole tree in Google Earth. # TODO
        file.write_xml({
          :Style => {
            :ListSyle => {
              :listItemType => "checkHideChildren"
            }
          }
        })

        # Text
        if @@options[:export_text]
          file.indent(:Folder){
            file.set(:name, 'Texts')
            @text.each_with_index{ |text, i|
              # Use no more than 30 characters for the name and the complete text as description.
              export_point(text.point, text.text[/^[^\n\r\:]{1,30}/], text.text)
            }
          }
        end

        # Construction Point
        if @@options[:export_cpoint]
          file.indent('Folder'){
            file.set(:name, 'Construction Points')
            @text.each_with_index{ |cpoint, i|
              export_point(cpoint.position, "Point #{i}", nil)
            }
          }
        end

        # Curve or Edges or extruded curve
        if @@options[:export_curve]
          case @@options[:curve_as]
          when :line_string
            @curve.each{ |data, i|
              data[:name] = "Curve #{i}" if data[:name].nil? || data[:name].empty?
              data[:extrude] = @@options[:hface_extrude] # TODO curve_extrude ?
              export_placemark_line_string(data)
            }
          when :track
            @curve.each_with_index{ |data, i|
              data[:name] = "Track #{i}" if data[:name].nil? || data[:name].empty?
              export_track(data)
            }
          end
        else
          @group.concat(@curve)
        end

        # Horizontal images or horizontal quadfaces
        if @@options[:export_himage_as_ground_overlay]
          @himage.each_with_index{ |ground_overlay, i|
            ground_overlay[:name] = "GroundOverlay #{i}" if ground_overlay[:name].nil? || ground_overlay[:name].empty?
            # Export it as GroundOverlay except if we want to ignore GE snapshots.
            unless ground_overlay[:name][/Google\sEarth\sSnapshot/] && @@options[:himage_exclude_gearth]
              export_ground_overlay(g)
            end
          }
        else
          @group.concat(@himage)
        end

        # Polygon or extruded polygon
        if @@options[:export_hface_as_polygon] == true
          @hface.each_with_index{ |polygon, i|
            polygon[:name] = "Polygon #{i}" if polygon[:name].nil? || polygon[:name].empty?
            polygon[:extrude] = @@options[:hface_extrude]
            export_placemark_polygon(polygon)
          }
        else
          @group.concat(@hface)
        end

        # Model (groups or components)
        if @@options[:export_group]
          dirs = {}
          @group.each_with_index{ |group, i|
            group[:name] = "Model #{i}" if group[:name].nil? || group[:name].empty?
            unless @@options[:model_exclude_gearth] && (group[:name][/Google\sEarth\sTerrain/] || group[:name][/Google\sEarth\sSnapshot/])
              case @@options[:group_as]
              when :model
                # Export model and remember model filepath in dirs{}
                begin
                  path = export_model(group)
                  # SketchUp saves textures in an accompanying folder with the same name. Remember that name.
                  dirs[path] = path.sub(/.dae$/, "")
                rescue Exception => e
                  key = @@options[:group_as]
                  log("An error occured when importing '#{key}' and the feature was skipped.", e.message, e.backtrace)
                  @failed[key] ||= 0
                  @failed[key] += 1
                else
                  @exported[key] ||= 0
                  @exported[key] += 1
                end
              when :multi_geometry # TODO
                export_multi_geometry(group)
              end
            end
          }
          Utilities.condense_collada(dirs) if @@options[:group_as] == :model && @@options[:model_condense_textures] == true && @group.length > 0 && dirs.length > 0
        end

        # Pages / scenes
        if @@options[:pages_as_tour] == true && @model.pages.count > 0
          export_tour(@@options[:page_as])
        # Single placemarks with Camera or LookAt
        elsif @@options[:export_page] == true
          @model.pages.each_with_index{ |page, i|
            page[:name] = "Scene #{i}" if page[:name].nil? || page[:name].empty?
            file.indent(:Placemark){
              file.set(:name, page[:name])
              export_time(page)
              case @@options[:page_as]
              when :camera
                export_camera(page)
              when :look_at
                export_look_at(page)
              end
            }
          }
        end

      } # Document
    } # kml
  } # XMLFile

  Dir.delete(@data_dir) unless Dir.entries(@data_dir).length > 2

  # Restore original selection
  @selection.each{ |s| @model.selection.add(s) }
end # def main



######################### export methods for entities  #########################



def export_model(hash)
  model = hash[:entity]
  name = hash[:name]
  description = hash[:description]

  if model.is_a?(Sketchup::Group)
    definition = model.entities.parent
  elsif model.is_a?(Sketchup::ComponentInstance)
    definition = model.definition
  elsif model.is_a?(Sketchup::Image)
    definition = Utilities::Image.get_definition(model)
  else # if entities within a group
    definition = model.parent
    model = hash[:parent] if hash[:parent]
  end
  transformation = model.transformation
  is_skewed = (
      transformation.xaxis.angle_between(transformation.yaxis).radians != 90 ||
      transformation.yaxis.angle_between(transformation.zaxis).radians != 90 ||
      transformation.xaxis.angle_between(transformation.zaxis).radians != 90
    ) # TODO: tolerance?

  # 1. Check if already exported.
  href = @exportedResources[definition] # returns href of existing file or nil
  # If it has not yet been exported or if it has a unique skewed transformation:
  if !href || is_skewed
    href = Utilities.to_filename(@data_dir, definition.name, ".dae")
    # Create an untransformed temporary copy of group/instance and select it.
    if is_skewed
      tmp_instance = @model.entities.add_group
      tmp_instance.entities.add_instance(definition, transformation)
      tmp_instance.entities[0].explode
      definition = tmp_instance.entities.parent
      transformation = tmp_instance.transformation
    else
      tmp_instance = @model.entities.add_instance(definition, Geom::Transformation.new)
    end
    # Select and export selection.
    @model.selection.clear
    @model.selection.add(tmp_instance)
    options_hash = {
      :triangulated_faces  => true,
      :doublesided_faces   => true,
      :edges               => false, # TODO: maybe optional?
      :materials_by_layer  => false,
      :author_attribution  => true,
      :texture_maps        => true,
      :selectionset_only   => true,
      :preserve_instancing => true # TODO: was false
    }
    success = @model.export(href, options_hash)
    raise(KMLExportError, "Model was not exported successfully.", [definition, href]) unless success
    @exportedResources[definition] = href
    # Delete temporary instance
    @model.entities.erase_entities(tmp_instance)
  end

  # 2. Calculate location and orientation of transformed group/instance.
  origin = transformation.origin
  origin.z = get_altitude(origin, @@options[:altitude_mode])
  location = @model.point_to_latlong(origin).to_a
  # Using an additional Geom::Transformation methods from dynamic components (after comment: old code)
  heading = transformation.rotz # 0-X_AXIS.angle_between(t.yaxis*Z_AXIS)*180/Math::PI.to_f # TODO
  tilt = -transformation.rotx # -90+t.yaxis.angle_between(Z_AXIS)*180/Math::PI.to_f
  roll = transformation.roty # t.zaxis.angle_between(Z_AXIS*t.yaxis)*180/Math::PI.to_f-90
  # Get the scale.
  scale = [1, 1, 1]
  if model.is_a?(Sketchup::Image)
    s = [model.width, model.height, 0]
    us = [definition.bounds.width, definition.bounds.height, 0]
  else
    s = model.scaled_size
    us = model.unscaled_size # = tmp_instance.unscaled_size
  end
  scale.each_with_index{ |x, i|
    scale[i] = (s[i]/us[i].to_f) unless s[i] == 0 || us[i] == 0
  }

  # 3. Print the KML.
  @file.write_xml({
    :Placemark => {
      :name => name,
      :description => description,
      Proc.new{ export_time(model) } => nil,
      :Model => {
        :altitudeMode => @@options[:altitude_mode],
        :Location => {
          :longitude => location.x,
          :latitude => location.y,
          :altitude => location.z
        },
        :Orientation => {
          :heading => heading,
          :tilt => tilt,
          :roll => roll
        },
        :Scale => {
          :x => scale[0],
          :y => scale[1],
          :z => scale[2]
        },
        :Link => {
          :href => Utilities.rel_path(href, File.dirname(@data_dir))
        }
      }
    }
  }) # write_xml
  return href
end # def export_model



def export_multi_geometry(hash) # multi_geometry with polygons and line_strings
  geom = hash[:entity]
  name = hash[:name]
  description = hash[:description]
  extrude = false
  if geom.is_a?(Sketchup::Group)
    definition = geom.entities.parent
  elsif geom.is_a?(Sketchup::ComponentInstance)
    definition = geom.definition
  else # if something else (entities within a group)
    definition = geom.parent
    geom = hash[:parent] if hash[:parent]
  end
  transformation = geom.transformation
  allfaces = {}
  edges = []
  definition.entities.each{ |entity| # TODO geom.entities?
    if entity.is_a?(Sketchup::Face)
      allfaces[entity.material] = [] if allfaces[entity.material].nil?
      allfaces[entity.material].push(entity)
    elsif entity.is_a?(Sketchup::Edge) && entity.faces.length == 0 && !entity.hidden?
      edges.push(entity) # orphan edges
    end
  }
  allfaces.each{|m, faces|
    facecolor = color_to_ge(faces[0].material || geom.material)
    linecolor = color_to_ge(faces[0].edges[0].material) || facecolor
    @file.write_xml({
      :Placemark => {
        :name => name,
        :description => description,
        Proc.new{ export_time(geom) } => nil,
        :Style => (
          if facecolor then {
            :LineStyle => {
              :color => linecolor,
              :colorMode => "normal",
              :width => 1
            },
            :PolyStyle => {
              :color => facecolor,
              :colorMode => "normal",
              :fill => 1,
              :outline => 1
            }
          }
          end
        ),
        :MultiGeometry => {
          Proc.new{
            faces.each{ |x|
              export_polygon(x, transformation, extrude, nil)
            }
            if !edges.empty?
              curves = Utilities.find_connected_edges(edges)
              curves.each{ |c|
                export_line_string(c, transformation, extrude, nil)
              }
            end
          } => nil
        }
      }
    }) # write_xml
  } # each
end # export_multi_geometry



def export_placemark_polygon(hash) # placemark with polygon
  faces = hash[:entities]
  name = hash[:name]
  parent = hash[:parent]
  transformation = (parent) ? parent.transformation : Geom::Transformation.identity
  faces.each{ |face|
    facecolor = color_to_ge(face.material || parent.material)
    linecolor = color_to_ge(face.edges[0].material) || facecolor
    @file.write_xml({
      :Placemark => {
        :name => name,
        Proc.new{ export_time(parent) } => nil,
        :Style => (
          if facecolor then {
            :LineStyle => {
              :color => linecolor,
              :colorMode => "normal",
              :width => 1
            },
            :PolyStyle => {
              :color => facecolor,
              :colorMode => "normal",
              :fill => 1,
              :outline => 1
            }
          }
          end
        ),
        Proc.new{ export_polygon(face, transformation, hash[:extrude]) } => nil
      }
    }) # write_xml
  } # each face
end # def export_placemark_line_string



def export_polygon(face, transformation, extrude=false)
  @file.write_xml({
    :Polygon => {
      :extrude => (extrude) ? 1 : 0,
      :tessellate => 0,
      # Extrusion should override clampToGround.
      :altitudeMode => (@@options[:altitude_mode] == :clampToGround && extrude) ? :relativeToGround : @@options[:altitude_mode],
      :outerBoundaryIs => {
        :LinearRing => {
          :coordinates => (
            outer = face.outer_loop.vertices.push(face.outer_loop.vertices.first) # to make loop closed
            extrusion = outer.inject(0){ |sum, v| sum + v.position.z - get_terrain_elevation(v.position) } / (outer.length|1) if @@options[:altitude_mode] == :relativeToGround && extrude
            outer.map{|v|
              p = v.position
              p.z = (@@options[:altitude_mode] == :relativeToGround && extrude) ? extrusion : get_altitude(p, @@options[:altitude_mode])
              p = @model.point_to_latlong(p.transform(transformation))
              p.to_a.join(",")
            }.join("\n")
          )
        }
      },
      :innerBoundaryIs => (if face.loops.length > 1 then
          face.loops.map{ |l|
            next if l.outer?
            {
              :LinearRing => {
                :coordinates => (
                  inner = l.vertices.push(l.vertices.first)
                  inner.map{ |v|
                    p = v.position
                    p.z = (@@options[:altitude_mode] == :relativeToGround && extrude) ? extrusion : get_altitude(p, @@options[:altitude_mode])
                    p = @model.point_to_latlong(p.transform(transformation))
                    p.to_a.join(",")
                  }.join("\n")
                )
              }
            }
          } # map
        end
      )
    }
  }) # write_xml
end # export_polygon



def export_placemark_line_string(hash) # placemark with line_string
  curves = hash[:entities]
  name = hash[:name]
  description = hash[:description]
  parent = hash[:parent] # may not necessarily exist
  linecolor = color_to_ge(curves[0].material || parent.material) # TODO: Can parent be nil?
  transformation = (parent.nil?) ? nil : parent.transformation
  curves = Utilities.find_connected_edges(curves)
  curves.each{ |curve|
    @file.write_xml({
      :Placemark => {
        :name => name,
        Proc.new{ export_time(parent) unless parent.nil? } => nil,
        :Style => (
          if linecolor then {
            :LineStyle => {
              :color => linecolor,
              :colorMode => "normal",
              :width => 1
            }
          } end
        ),
        Proc.new{ export_line_string(curve, transformation, hash[:extrude]) } => nil
      }
    }) # write_xml
  }
end # def export_placemark_line_string



def export_line_string(curve, transformation, extrude=false)
  extrusion = curve.inject(0){ |sum, v| v.position.z - get_terrain_elevation(v.position) } / (curve.length|1) if @@options[:altitude_mode] == :relativeToGround && extrude
  @file.write_xml({
    :LineString => {
      :extrude => (extrude) ? 1 : 0,
      :tessellate => 0,
      # Extrusion should override clampToGround.
      :altitudeMode => (@@options[:altitude_mode] == :clampToGround && extrude) ? :relativeToGround : @@options[:altitude_mode],
      :coordinates => curve.map{ |v|
        p = v.position
        p.transform!(transformation) if transformation
        p.z = (@@options[:altitude_mode] == :relativeToGround && extrude) ? extrusion : get_altitude(p, @@options[:altitude_mode])
        @model.point_to_latlong(p).to_a.join(",")
      }.join(" ")
    }
  }) # write_xml
end # def export_line_string



def export_track(hash)
  name = hash[:name]
  description = hash[:description]
  ents = hash[:entities]
  parent = hash[:parent] # may not necessarily exist
  material = (!ents.nil?)? ents[0].material : parent.material
  linecolor = color_to_ge(material)
  transformation = (parent.nil?) ? nil : parent.transformation
  curves = Utilities.find_connected_edges(ents)
  curves.each_with_index{ |c, i|
    @file.write_xml({
      :Placemark => {
        :name => "#{name} #{i}",
        Proc.new{ export_time(parent) unless parent.nil? } => nil,
        :Style => (
          if linecolor then {
            :LineStyle => {
              :color => linecolor,
              :colorMode => "normal",
              :width => 1
            },
            :IconStyle => {
              :scale => 1,
              :color => linecolor
            }
          } end
        ),
        :"gx:Track" => {
          :altitudeMode => @@options[:altitude_mode],
          :"gx:coord" => (
            c.collect{ |v|
              p = v.position
              p.transform!(transformation) if transformation
              p.z = get_altitude(p, @@options[:altitude_mode])
              string = @model.point_to_latlong(p).to_a.join(" ")
            }
          ),
          :when => (
            unless parent.nil?
              time_begin = parent.get_attribute("Time", "begin")
              time_end = parent.get_attribute("Time", "end")
              unless time_begin.nil? || time_end.nil?
                [time_begin, time_end]
              end
            end
          ),
          :"gx:angles" => (
            (c.length-1).times{ |i| # TODO: map_with_index?
              p = c[i].position
              q = c[i+1].position
              dx = p.x-q.x # relative changes between two vertices
              dy = p.y-q.y
              dz = p.z-q.z
              heading = -Math.atan2(dy,dx).radians - 180
              dxy = Math.sqrt(dx*dx+dy*dy)
              roll = -Math.atan2(dz,dxy).radians
            }.
            # Add one dummy element to match the number of vertices).
            push("#{heading} 0 #{roll}")
          ),
        }
      }
    }) # write_xml
  } # curves.each
end # def export_track



def export_point(position, name, description)
  @file.write_xml({
    :Placemark => {
      :name => "#{name} #{i}",
      :description => ("<![CDATA[#{description}]]>" unless description.nil? || description.empty?),
      :Point => {
        :altitudeMode => @@options[:altitude_mode],
        :coordinates => (
          position.z = get_altitude(position, @@options[:altitude_mode])
          p = @model.point_to_latlong(position)
          p.to_a.map{ |c| c.to_f }.join(",")
        )
      }
    }
  }) # write_xml
end # def export_point



def export_ground_overlay(hash)
  image = hash[:entity]
  name = hash[:name]
  description = hash[:description]
  tw = Sketchup.create_texture_writer

  if image.is_a?(Sketchup::Image)
    definition = Utilities::Image.get_definition(image)
    face = definition.entities.find{|x| x.is_a?(Sketchup::Face)} # TODO: or use grep?
  else # textured quadface
    face = image # (to make the following more logical)
  end
  # create temporary group to export texture image without uv-map
  tmp_group = @model.entities.add_group
  tmp_group.material = face.material
  tw.load(tmp_group)
  image_file_name = File.basename(face.material.texture.filename)
  href = @exportedResources[face.material]
  unless href # export as a new file
    href = Utilities.to_filename(@data_dir, image_file_name)
    tw.write(tmp_group, href, true)
    @exportedResources[face.material] = href
  end
  @model.entities.erase_entities(tmp_group)

  @file.write_xml({
    :GroundOverlay => {
      :name => name,
      :description => ("<![CDATA[#{description}]]>" unless description.nil? || description.empty?),
      :Icon => {
        :href => Utilities.rel_path(href, File.dirname(@data_dir))
      },
      Proc.new{
        if image.is_a?(Sketchup::Image)
          # no skewed transformation of Sketchup::Image supported
          o = image.origin
          w = image.width
          h = image.height
          rot = image.zrotation
          c = image.bounds.center
          northwest = @model.point_to_latlong( c+[-0.5*w,0.5*h,0] )
          southeast = @model.point_to_latlong( c+[0.5*w,-0.5*h,0] )
          {
            :LatLonBox => {
              :north => northwest.y.to_f,
              :south => southeast.y.to_f,
              :east => southeast.x.to_f,
              :west => northwest.x.to_f,
              :rotation => rot*180/Math::PI
            }
          }
        else
          transformation = (hash[:parent]) ? hash[:parent].transformation : Geom::Transformation.new
          c = face.outer_loop.vertices # counter-clockwise array of vertices
          c.reverse if face.normal == [0,0,-1]
          c.collect!{|x| x = @model.point_to_latlong(x.position.transform(transformation)).to_a.join(",")}
          {
            :"gx:LatLonQuad" => {
              :coordinates => c.join(" ")
            }
          }
        end # if
      } => nil # Proc
    }
  }) # write_xml
end # def export_ground_overlay



def export_camera(page)
  camera = page.camera
  vector = (camera.target - camera.eye).normalize
  # eye
  eye = camera.eye
  eye.z = get_altitude(eye, @@options[:altitude_mode])
  eye = @model.point_to_latlong(eye)
  # heading
  vectorxy = [vector.x,vector.y,0].normalize
  heading = Math.asin(vectorxy.x).radians.to_f
  heading = 180 - heading.to_f if vectorxy.y < 0
  # tilt
  tilt = 90 + Math.asin(vector.z).radians
  # roll
  roll = (vector*camera.up).angle_between([0,0,1]).radians - 90
  # fov / field of view not possible in kml

  @file.write_xml({
    :Camera => {
      :longitude => eye.x.to_f,
      :latitude => eye.y.to_f,
      :altitude => eye.z.to_f,
      :heading => heading,
      :tilt => tilt,
      :roll => roll,
      :altitudeMode => @@options[:altitude_mode],
      :"gx:ViewerOptions" => '<option name="groundnavigation" type="boolean">0</option>'
    }
  }) # write_xml
end # def export_camera



def export_look_at(page)
  camera = page.camera
  vector = (camera.target - camera.eye).normalize
  # heading
  vectorxy = [vector.x, vector.y, 0].normalize
  heading = Math.asin(vectorxy.x).radians.to_f
  heading = 180-heading.to_f if vectorxy.y < 0
  # tilt
  tilt = 90 + Math.asin(vector.z).radians
  # range
  range = (camera.target - camera.eye).length.to_m
  # target
  t = camera.target
  if t.z < 0 && vector.z != 0
    vector.length = -camera.eye.z / vector.z.to_f
    t = camera.eye + vector
    range = vector.length.to_f.to_m
  end
  t.z = get_altitude(t, @@options[:altitude_mode])
  target = @model.point_to_latlong(t)
  target.z = get_altitude(target, @@options[:altitude_mode])
  # fov / field of view not possible in kml

  @file.write_xml({
    :LookAt => {
      :longitude => target.x.to_f,
      :latitude => target.y.to_f,
      :altitude => target.z.to_f,
      :heading => heading,
      :tilt => tilt,
      :range => range,
      :altitudeMode => @@options[:altitude_mode],
      :"gx:ViewerOptions" => '<option name="groundnavigation" type="boolean">0</option>'
    }
  }) # write_xml
end # def export_look_at



def export_tour(page_as)
  @file.write_xml({
    :"gx:Tour" => {
      :name => "Tour",
      :"gx:Playlist" => {
        Proc.new{
          @model.pages.map{ |page|
            {
              :"gx:FlyTo" => {
                :"gx:duration" => ((page.transition_time != -1) ? page.transition_time : 1),
                :"gx:flyToMode" => "smooth",
                Proc.new{
                  if page_as == :camera
                    export_camera(page)
                  elsif page_as == :look_at
                    export_look_at(page)
                  end
                } => nil # Proc
              },
              :"gx:Wait" => (
                if p.delay_time != -1 then
                  { :"gx:duration" => p.delay_time }
                end
              )
            }
          } # map page
        } => nil # Proc
      }
    }
  }) # write_xml
end # def export_tour



def export_time(entity)
  time_begin = entity.get_attribute("Time","begin")
  time_end = entity.get_attribute("Time","end")
  time_when = entity.get_attribute("Time","when")

  if !time_begin.nil? || !time_end.nil?
    @file.write_xml({
      :TimeSpan => {
        :begin => (time_begin unless time_begin.nil?),
        :end => (time_end unless time_end.nil?)
      }
    }) # write_xml
  elsif !time_when.nil?
    @file.write_xml({
      :TimeStamp => {
        :when => time_when
      }
    }) # write_xml
  end
end # def export_time



def export_style(entity) # TODO
end # def export_style



#########################       Helper methods        ##########################



# Converts SketchUp's rgba colors into hexadecimal aabbggrr.
# @param [Sketchup::Material,Sketchup::Color] color
# @returns [String] a color in KML's color format
#
def color_to_ge(color)
  return nil unless color.is_a?(Sketchup::Color) || color.is_a?(Sketchup::Material)
  # Get the alpha value from the material if possible, because Sketchup::Color's alpha is somehow opaque.
  if color.is_a?(Sketchup::Material)
    alpha = (color.alpha.to_f*255).to_i
    color = color.color
    color.alpha = alpha
  end
  # SKP color order is  r, g, b, a.
  skp_color_array = color.to_a
  # KML color order is  aabbggrr
  kml_color_array = skp_color_array.reverse
  kml_color = sprintf("%02x%02x%02x%02x", *kml_color_array)
  return kml_color
end # def color_to_ge



# Calculate altitude of a point, considering altitudeMode.
# (everything in model coordinates)
# @param [Geom::Point3d,Array<Numeric>] point in the model
# @param [Symbol] mode
# @returns [Float] the altitude according to mode
#
def get_altitude(point, mode)
  height = point.z.to_f
  if mode == :clampToGround || mode.nil?
    ter = get_terrain_elevation(point)
    return 0
  elsif mode == :relativeToGround
    ter = get_terrain_elevation(point)
    return (height - ter)
  elsif mode == :absolute
    return height + @elevation.m
  else # TODO: relativeToSeaFloor, clampToSeaFloor (requires gx:altitudeMode)
    return height + @elevation.m
  end
end # def get_altitude



# Get the elevation of the Google Earth Terrain (in model coordinates)
# @param [Geom::Point3d,Array<Numeric>] point
# @returns [Float] the elevation at the given point
#
def get_terrain_elevation(point)
  p = Geom::Point3d.new(point)
  p.z = @model.bounds.min.z
  # Raytest:
  # Do not use [ray]=[point,vector] but point,vector,wysiwyg_flag
  # because SU<8M1 interprets two parameters also as a ray
  raytest = @model.raytest(p, Z_AXIS, false)
  until raytest.nil? || raytest[1][0].is_a?(Sketchup::Group) && raytest[1][0].name[/Google Earth Terrain/]
      p.z = raytest[0].z
      raytest = @model.raytest(p, Z_AXIS, false)
  end
  return (!raytest.nil?) ? raytest[0].z : 0
end # def get_terrain_elevation



def log(*message)
  # file_line = caller.first.to_s[/[^\/\:]+\:\d+(?=\:)/]
  # message.first << " (#{file_line})"
  @log << message
end



def raise(error, reason, debug_data)
  file_line = caller.first.to_s[/[^\/\:]+\:\d+(?=\:)/]
  Kernel.raise(error, "#{reason} (#{file_line})", debug_data)
end



class KMLExportError < StandardError
end



end # class Exporter



class XMLFile < File
  @@indent = 2
  def initialize(*args)
    @indentation = 0
    super(*args)
  end

  def indent(tag_string=nil, &block)
    # The tag_string can contain attributes.
    tag = tag_string.to_s[/^\S+/] if tag_string
    puts("<#{tag_string}>") if tag_string
    @indentation += 1
    block.call
    @indentation -= 1
    puts("</#{tag}>") if tag_string
  end

  def puts(*args)
    i = @@indent * @indentation
    args.map!{ |arg| arg.to_s.rjust(i + arg.length) }
    super(*args)
  end

  def set(property, value)
    puts("<#{property}>#{value}</#{property}>")
  end

  def set_cdata(property, value)
    puts("<#{property}><![CDATA[#{value}]]></#{property}>")
  end

  def write_xml(*array)
    array.each{ |a|
      next unless a.is_a?(Hash)
      a.each{ |k, v|
        if v.is_a?(Hash)
          indent(k){
            write_xml(v)
          }
        elsif k.is_a?(Method) || k.is_a?(Proc)
          if v.is_a?(Array)
            k.call(*v)
          else
            k.call(v)
          end
        elsif !v.nil?
          v = v.to_s
          if v[/\n/]
            # indent(k){ puts(v) }
            puts("<#{k}>\n#{v}\n#{" " * @@indent * @indentation}</#{k}>")
          else
            puts("<#{k}>#{v}</#{k}>")
          end
        end
      }
    }
  end

end # class XMLFile



module Utilities



# Filter edges that are connected and sort them.
# @param [Array<Sketchup::Edge>] edges
# @returns [Array<Array<Sketchup::Vertex>>] array(s) of sorted vertices
#
def self.find_connected_edges(edges)
  counter = 0 # Count remaining edges to compare.
  curves = []
  sorted = []
  while edges.length > 0
    if counter == 0
      sorted = edges.shift.vertices # Start new sorted array.
      counter = edges.length
    end

    if edges.length > 0
      edge = edges.shift # Take another edge and compare.
      ev = edge.vertices
      o = *(ev-sorted) # Find vertex that is not yet in in sorted sorted array.
      if ev.include?(sorted.first) # If edges shares a vertex with sorted array sorted, add the other vertex.
        o = sorted.last if o.nil? # Only for closed rings.
        sorted.unshift(o)
        counter = edges.length
      elsif ev.include?(sorted.last)
        o = sorted.first if o.nil? # Only for closed rings.
        sorted.push(o)
        counter = edges.length
      # If it is not connected to the previous edge, try the next edge.
      else
        edges.push(edge)
        counter -= 1
      end
    end
    curves.push(sorted) if counter == 0
  end # while
  return curves
end # def find_connected_edges



# Convert a random string into a proper, unique filename.
# @param [String] dir  directory where the file should go
# @param [String] string  any user input
# @param [String] ext  optionally force a specific extension if the string doesn't contain one
# @return [String] a complete file path
#
def self.to_filename(dir, string, ext=nil)
  # Separate extension from basename.
  ext = ::File.extname(string) if ext.nil?
  string = ::File.basename(string, ext.to_s)
  # Clean up basename.
  string = string[/.{1,30}/].gsub(/\n|\r/,"").gsub(/[^0-9a-zA-Z\-\_\.\(\)]+/,"_").gsub(/^_+|_+$/,"").to_s
  string = "file" if string.empty?
  base = ::File.join(dir, string)
  # Detect collision of filenames and return alternative filename (numbered).
  base_orig = base
  i = 0
  while ::File.exists?(base + ext)
    base = base_orig + i.to_s
    i += 1
  end
  return base + ext
end # def to_filename



# Convert absolute file path into relative path, relative to
# - <base> directory or
# - next common directory with <file> given as second argument
# @param [String] abs  and absolute file path
# @param [String] base  a parent directory of the absolute file path
# @returns [string] a relative file path
#
def self.rel_path(abs, base)
  base = base.sub(/\/?$/,"/")
  i = 0
  i += 1 until abs[i] != base[i]
  i -= 1 if i>0
  rel = abs.scan(/^.{0,#{i}}\/(.*$)/).to_s
  return (i==0)? abs : rel
end # def rel_path



# Check if <directory> contains a file that is identical to <file>
# @param [String] file
# @param [String] directory
# @returns [String] filename
#
def self.file_duplicate?(file, directory)
  Dir.entries(directory).find{ |f|
    ::File.size(::File.join(directory, f)) == ::File.size(file)
  }
end # def file_duplicate?



# Each exported collada file has its own ressources folder.
# → Replace the many folders by a single "textures" folder that contains all ressources except duplicates.
# → Make all collada files refer to the new folder.
# @param [Hash] dirs  a hash assigning each collada file path a corresponding ressources folder
#
def self.condense_collada(dirs)
  counter = 0 # counter for generic new filenames
  tex_dir = File.join(File.dirname(dirs.keys.first), "textures")
  Dir.mkdir(tex_dir) unless ::File.exist?(tex_dir)
  dirs.each{ |c, d| # c = collada file; d = directory corresponding to collada file
    fileNames = Hash.new # hash to remember file name assignment
    Dir.foreach(d){ |f|
      next if f == "." || f == ".."
      # check if file f is already contained in tex_dir, returns name o existing file
      newFileName = self.file_duplicate?(::File.join(d, f), tex_dir)
      # if file is not yet in tex_dir, move it with new generic filename
      if newFileName == nil
        extension = f[/\.\w{0,3}$/]
        newFileName = "texture#{counter}#{extension}"
        Utilities::File.move(::File.join(d, f), ::File.join(tex_dir, newFileName))
        counter += 1
        # otherwise delete it since it is already in tex_dir
      else
        ::File.delete(d+f)
      end
      relSource = self.rel_path(::File.join(d, f), c)
      relTarget = self.rel_path(::File.join(tex_dir, newFileName), c)
      # remember filename assignment
      fileNames[relSource] = relTarget
    }
    Dir.foreach(d){ |f|
      next if f == '.' || f == '..'
      ::File.delete(::File.join(d, f))
    } if Dir.entries(d)[2]
    Dir.delete(d)
    # read Collada file and replace filenames
    collada = IO.read(c)
    fileNames.each{ |old, new|
      collada.gsub!(old, new)
    }
    colladaFile = ::File.open(c, "w")
    colladaFile.write(collada)
    colladaFile.close
  }
  return true
rescue Exception => e
  $stderr.write("Error #{e.message}")
  $stderr.write(e.backtrace.join)
end # def condense_collada



module File


def self.move(from, to, verbose=false)
  #to = ::File.catname(from, to)
  if RUBY_PLATFORM =~ /djgpp|(cyg|ms|bcc)win|mingw/ && ::File.file?(to)
    ::File.unlink(to)
  end
  fstat = ::File.stat(from)
  begin
    ::File.rename(from, to)
  rescue
    begin
      ::File.symlink(::File.readlink(from), to) && ::File.unlink(from)
    rescue
      from_stat = ::File.stat(from)
      ::File.syscopy(from, to) && ::File.unlink(from)
      ::File.utime(from_stat.atime, from_stat.mtime, to)
      begin
        ::File.chown(fstat.uid, fstat.gid, to)
      rescue
      end
    end
  end
end

end # module File



module Image

   def self.get_definition(image)
      return image.model.definitions.find{ |d|
        d.image? && d.instances.include?(image)
      }
   end

   def self.get_transformation(image)
      origin = image.origin
      axes = image.normal.axes
      tr = Geom::Transformation.axes(ORIGIN, axes[0], axes[1], axes[2])
      tr = tr*Geom::Transformation.rotation(ORIGIN, Z_AXIS, image.zrotation)
      tr = (tr*Geom::Transformation.scaling(ORIGIN, image.width/image.pixelwidth, image.height/image.pixelheight, 1)).to_a
      tr[12] = origin.x
      tr[13] = origin.y
      tr[14] = origin.z
      return Geom::Transformation.new(tr)
   end

   def self.set_transformation(image, tr)
      image.transform!(image.transformation.inverse*tr)
   end

end # module Image



end # module Utilities



unless file_loaded?(__FILE__)
  # Sketchup.register_exporter(Exporter.new) # Hey SketchUp 2014!

  # Command for KML Export
  cmd_kmlexporter = UI::Command.new(TRANSLATE["Export KML File"]) {
    begin
      exporter = Exporter.new
      dialog = exporter.do_options
      block = Proc.new{
        suggested_name = File.basename(Sketchup.active_model.path).sub(/\.skp$/, ".kml")
        suggested_path = File.join(ENV['USERPROFILE']||ENV['HOME']||Dir.pwd, suggested_name)
        path = UI.savepanel(TRANSLATE["Save KML File"])
        next unless path # If file dialog has been closed/cancelled.
        success = exporter.write_file(path)
        UI.messagebox(TRANSLATE["Export failed."], MB_OK) if success == 4
      }
      # Attach an event handler to when the dialog is closed.
      if dialog.visible?
        dialog.on_close(&block)
      # If it is a modal dialog (SketchUp on Windows), it has halted execution
      # until it was closed, so we can directly run the exporter.
      else
        block.call
      end
    rescue Exception => e
      $stderr.write(e.message)
      $stderr.write(e.backtrace.join("\n"))
    end
  }
  cmd_kmlexporter.tooltip = TRANSLATE["Exports geo-referenced data as KML file."]
  cmd_kmlexporter.status_bar_text = TRANSLATE["Select entities to export or export the whole model. In the options dialog, choose to which KML features the entities should be written."]
  cmd_kmlexporter.small_icon = File.join(DIR, "images", "icon_kmlexport_16.png")
  cmd_kmlexporter.large_icon = File.join(DIR, "images", "icon_kmlexport_24.png")
  KMLTools::MENU = UI.menu("Plugins").add_submenu("KML Tools") unless defined?(self::MENU)
  KMLTools::MENU.add_item(cmd_kmlexporter)
  file_loaded(__FILE__)
end



end # module KMLTools



end # module AE
