=begin

Copyright 2011-2013, Andreas Eisenbarth
All Rights Reserved

Permission to use, copy, modify, and distribute this software for
any purpose and without fee is hereby granted, provided that the above
copyright notice appear in all copies.

THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

Name:         KMLImporter.rb
Author:       Andreas Eisenbarth
Usage:        (menu) File → Import → Keyhole Markup Language (.kml)
              (menu) Plugins → KML Tools → Import KML File
Description:  It imports KML Files (Keyhole Markup Language) containing
              geo-referenced data.
              Extract kmz files (zip-archives) first. All ressources must be
              available locally, that means network links won't work.
              Supports: <Model>, <Multigeometry>, <Polygon>, <LineString>, <gx:Track>,
              <Point>, <GroundOverlay>, <PhotoOverlay>, <Camera>, <LookAt>, <gx:Tour>

              to be added: <gx:MultiTrack>,
              <gx:AnimatedUpdate>? (only AbstractViews)
Version:      2.0.0-dev
Date:         31.05.2013


TODO:
• quiet mode for command line import?
• import_geometry, import_views, import_folder not yet respected
• Track: import angles and time as attributes
• Problem with \n\t
• ExtendedData not yet supported
• export of "Folder" groups not yet supported
=end
require 'sketchup.rb'



module AE



module KMLTools



DIR = File.dirname(__FILE__) unless defined?(self::DIR)
require(File.join(DIR, "Translate.rb"))
TRANSLATE = Translate.new("KMLTools", File.join(DIR, "lang")) unless defined?(self::TRANSLATE)
require(File.join(DIR, "Dialog.rb"))
require(File.join(DIR, "Options.rb"))
@debug ||= true
class << self
  attr_accessor :debug
end



class Importer < Sketchup::Importer

  IMPORT_SUCCESS                        = 0
  IMPORT_FAILED                         = 1
  IMPORT_CANCELLED                      = 2
  IMPORT_FILE_NOT_FOUND                 = 4
  IMPORT_SKETCHUP_VERSION_NOT_SUPPORTED = 5

  KML_ATTRIBUTES = "KML Attributes"

  @@options ||= Options.new("KMLImporter", {
    :show_options          => false,
    :show_report           => true, # TODO: false
    :geometry_grouped      => true,
    :import_geometry       => true,
    :import_line_string    => true,
    :import_model          => true,
    :import_point          => true,
    :import_polygon        => true,
    :import_track          => true,
    :import_ground_overlay => true,
    :import_photo_overlay  => true,
    :import_camera         => true,
    :import_look_at        => true,
    # Or rather:
    :import_folder         => true,
    :import_views          => true,
    :point_as              => :cpoint, # :cpoint, :text
  })

  def initialize(options={}) # TODO: since we are asynchronous, we need :on_success, :on_failure
    @options_dlg = nil
    @@options.update(options)
    @quiet = (options[:quiet] == true)
    @xml_bridge = nil
    @log = []
    # A list of features and how many of them were imported.
    @failed = {}
    @imported = {}
    @model = nil
    # Absolute elevation above sea level of origin, in meter.
    @elevation = 0
    # Holds all global styles that are identified by id.
    @styles = Hash.new({:linecolor => nil, :facecolor => nil, :facefill => 1})
  end

  def reset
    @log.clear
    @failed.clear
    @imported.clear
    @model = Sketchup.active_model
    @elevation = 0
    @styles.clear
  end

  def description
    return "Keyhole Markup Language (*.kml)"
  end

  def file_extension
    return "kml"
  end

  def id
    return "com.sketchup.exporters.ae.kml"
  end

  def supports_options?
    return true
  end

  def do_options
    begin
      show_options
    rescue Exception => e
      $stderr.write(e.message)
      $stderr.write(e.backtrace.join("\n"))
    end
  end

  # This method is called by SketchUp after the user has selected a file
  # to import. This is where you do the real work of opening and
  # processing the file.
  def load_file(path, status=true)
    reset
    # Validate the file path.
    return IMPORT_FILE_NOT_FOUND unless path.is_a?(String)
    file = File.expand_path(path)
    return IMPORT_FILE_NOT_FOUND unless File.exists?(path)
    # Read the file.
    success = main(path)
    Sketchup.send_action("viewZoomExtents:")
    return IMPORT_SUCCESS
  rescue Exception => e
    $stderr.write(e.message)
    $stderr.write(e.backtrace.join("\n"))
    return IMPORT_FAILED
  end



private



# Ask the user which categories to import (and other options).
#
def show_options
  @options_dlg = Dialog.new(TRANSLATE["Import KML – Options"], false, "AE_KMLImporterOptions", 550, 400, 300, 300, true)
  html_path = File.join(DIR, "html", "KMLImporterOptions.html")
  @options_dlg.set_file(html_path)

  # Callbacks.

  @options_dlg.on_show{ |dlg, param|
    TRANSLATE.webdialog(dlg)
    dlg.call_function("AE.KMLImporterOptions.initialize", @@options.get_all)
  }

  @options_dlg.on("update_options") { |dlg, options|
    @@options.update(options)
  }

  @options_dlg.show_modal
  return @options_dlg
end # def options_dialog



def main(path)

  @file_dir = File.dirname(path)
  kml_string = IO.read(path)

  # Get elevation at the model's origin.
  # If the model is geo-referenced, look whether it has elevation info.
  if (@model.respond_to?(:georeferenced?) && @model.georeferenced? ) ||
        !@model.get_attribute("GeoReference", "Longitude").nil? # SU7.1+
    elevation = @model.get_attribute("GeoReference", "ModelTranslationZ")
    @elevation = -elevation.to_m if elevation

  # Otherwise try to geo-reference the model
  else
    coordinate_string = kml_string[/<longitude>([e0-9\.-]+)<\/longitude>(?:\s|\r|\n)*<latitude>([e0-9\.-]+)<\/latitude>(?:\s|\r|\n)*<altitude>([e0-9\.-]+)<\/altitude>/]
    coordinate_string = kml_string[/<coordinates>(?:\s|\r|\n)*([e0-9\.-]+),([e0-9\.-]+),?([e0-9\.-]+)?/] unless coordinate_string
    if coordinate_string
      coordinates = coordinate_string.scan(/[0-9\.-][e0-9\.-]*/).map{ |c| c.to_f }
      set_geo_reference(*coordinates) if coordinates.length >= 2
    else
      unless @quiet
        UI.messagebox(TRANSLATE["The model needs to be geo-referenced first."])
        Sketchup.send_action("getCurrentView:") unless OSX
      end
      return IMPORT_CANCELLED
    end
  end

  # Check if there is a Google Earth terrain group, otherwise offer to import one.
  unless @quiet || @model.entities.find{ |e| e.is_a?(Sketchup::Group) && e.name == "Google Earth Terrain" }
    result = UI.messagebox(TRANSLATE["The model does not contain a Google Earth terrain. Do you want to import one first?"], MB_YESNO)
    Sketchup.send_action("getCurrentView:") if result == 6 unless OSX # Which gives an error about unsupported string argument.
    return IMPORT_CANCELLED
  end

  # WebDialog to convert XML into a Ruby hash.
  @xml_bridge = Dialog.new(TRANSLATE["Import KML file"], false, "AE_KMLImporter", 200, 120, 300, 300, true)
  html_path = File.join(DIR, "html", "KMLImporter.html")
  @xml_bridge.set_file(html_path)
  # Callbacks.
  @xml_bridge.on_show{ |dlg, param|
    TRANSLATE.webdialog(dlg)
    dlg.call_function("AE.KMLImporter.initialize")
    # If needed, get the absolute elevation either from the model or the imported file.
    if @elevation.nil? || @elevation == 0
      lon, lat, alt = @model.point_to_latlong(ORIGIN).to_a
      dlg.call_function("AE.KMLImporter.getElevation", lat, lon)
    end
    # Load the file, let the browser's xml parser read it and convert it to a ruby hash.
    dlg.call_function("AE.KMLImporter.loadFile", kml_string)
  }

  @xml_bridge.on("receive_elevation") { |dlg, elevation|
    if elevation.is_a?(Numeric) && elevation != 9999
      @elevation = elevation
    end
  }

  @xml_bridge.on("receive_xml") { |dlg, xml_hash|
    if Sketchup.version_number > 7000000
      @model.start_operation(TRANSLATE["Import KML File"], true) # SU7.0
    else
      @model.start_operation(TRANSLATE["Import KML File"])
    end
    begin
      # Import the file.
      walk_tree(xml_hash, {})
    rescue Exception => e
      @model.abort_operation
      raise
    end
    (@imported.empty?) ? @model.abort_operation : @model.commit_operation
    if @@options[:show_report]
      # Show a summary of imported features.
      @xml_bridge.call_function("AE.KMLImporter.show_report", @imported, @failed, @log)
    else
      @xml_bridge.close
    end
  }

  @xml_bridge.show_modal
end # def main



@@import_methods = {
  # Container
  :Document      => :walk_tree,
  :Folder        => :import_folder,
  # StyleSelector
  :Placemark     => :import_placemark,
  :Style         => :import_style,
  # Geometry
  :LineString    => :import_line_string,
  :Model         => :import_model,
  :MultiGeometry => :walk_tree,
  :Point         => :import_point,
  :Polygon       => :import_polygon,
  :Track         => :import_track,
  # Overlay
  :GroundOverlay => :import_ground_overlay,
  :PhotoOverlay  => :import_photo_overlay,
  # AbstractView
  :Camera        => :import_camera,
  :LookAt        => :import_look_at,
  :"gx:Tour"     => :import_tour
}



@@properties = [
  :name,
  :visibility,
  :open,
  :"atom:author",
  :"atom:name",
  :"atom:link",
  :address,
  :Snippet,
  :description,
  :styleUrl,
  :@attributes,
  :TimeStamp,
  :TimeSpan
]



def walk_tree(obj, inherited={})
  return unless obj.is_a?(Hash)
  # Array of created entities.
  entities = []
  # Import Style first
  if obj[:Style].is_a?(Hash) && obj.include?(:Style)
    key = :Style
    value = obj[:Style]
    begin
      import_style(value, inherited)
    rescue Exception => e
      log("An error occured when importing '#{key}' and the feature was skipped.", e.message, e.backtrace)
      @failed[key] ||= 0
      @failed[key] += 1
    else
      @imported[key] ||= 0
      @imported[key] += 1
    end
    obj.delete(key)
  end
  # Loop over all Features
  obj.each{ |key, value|
    next if @@properties.include?(key)
    #
    if @@import_methods.include?(key)
      method = method(@@import_methods[key])
      # If parent contains many elements for this key, they are in an array:
      if value.is_a?(Array)
        value.each{ |v|
          begin
            entities << method.call(v, inherited)
          rescue Exception => e
            log("An error occured when importing '#{key}' and the feature was skipped.", e.message, e.backtrace)
            @failed[key] ||= 0
            @failed[key] += 1
          else
            @imported[key] ||= 0
            @imported[key] += 1
          end
        }
        obj.delete(key)
      # or if only one element exists for this key:
      else
        begin
          entities << method.call(value, inherited)
        rescue Exception => e
          log("An error occured when importing '#{key}' and the feature was skipped.", e.message, e.backtrace)
          @failed[key] ||= 0
          @failed[key] += 1
        else
          @imported[key] ||= 0
          @imported[key] += 1
        end
        obj.delete(key)
      end
    # Not supported feature
    elsif (65..90).include?(key.to_s[0]) # Uppercase?
      @failed[key] ||= 0
      @failed[key] += 1
    end
  }
  # Return an array of created entities.
  return entities.compact
end



######################### import methods for entities  #########################



def set_attributes(entity, data, type=nil)
  @@properties.each{ |p|
    next unless data.include?(p)
    if p == :@attributes
      if data[p].include?(:id)
        entity.set_attribute(KML_ATTRIBUTES, "id", data[p][:id])
      end
      next
    end
    #if !data[p].is_a?(Hash)
      entity.set_attribute(KML_ATTRIBUTES, p.to_s, data[p])
    #else
    #  data[p].each{ |k, v|
    #    entity.set_attribute(KML_ATTRIBUTES, "#{p}.#{k}", v)
    #  }
    #end
  }
  entity.set_attribute(KML_ATTRIBUTES, "type", type) if type
end



def import_style(data, inherited={})
  return {} if data.nil? || data.empty?
  id = data[:@attributes][:id] if data[:@attributes]
  style = {}

  if data[:LineStyle]
    linecolor = data[:LineStyle][:color]
    style[:linecolor] = color_to_su(linecolor) unless linecolor.nil? || linecolor.empty?
  end

  if data[:PolyStyle]
    facecolor = data[:PolyStyle][:color]
    style[:facecolor] = color_to_su(facecolor) unless facecolor.nil? || facecolor.empty?
    facefill = data[:PolyStyle][:fill]
    style[:facefill] = facefill.to_i unless facefill.nil? || facefill.empty?
  end

  @styles[id] = style if id
  return style
end # def import_style



def import_time(data, entity)
  # Apply data.
  if data[:TimeSpan]
    entity.set_attribute("Time", "begin", data[:TimeSpan][:begin]) if data[:TimeSpan][:begin]
    entity.set_attribute("Time", "end", data[:TimeSpan][:end]) if data[:TimeSpan][:end]
  elsif data[:TimeStamp]
    entity.set_attribute("Time", "when", data[:TimeStamp][:when]) if data[:TimeStamp][:when]
  end
end # def import_time



def import_folder(data, inherited={})
  name = data[:name] || TRANSLATE["Folder %0", @model.layers.length+1]
  # Create a layer.
  layer = @model.layers.add(name)
  layer.visible = (data[:visibility].to_i == 1) if data[:visibility]
  inherited[:current_layer] = layer
  # Create a group.
  group = (inherited[:current_context] || @model).entities.add_group
  group.name = name
  group.description = data[:description] if data[:description]
  group.layer = layer
  inherited[:current_context] = group
  # Add Features.
  set_attributes(layer, data, "Folder")
  set_attributes(group, data, "Folder")
  entities = walk_tree(data, inherited)
  return entities
end



def import_placemark(data, inherited={})
  # Collect all data.
  inherited[:name] = data[:name]
  inherited[:description] = data[:description]
  # Get the style.
  inherited[:style] = {} # TODO: or ||=
  if data[:styleUrl]
    id = data[:styleUrl].to_s.sub(/^\#/, "")
    inherited[:style].merge!(@styles[id]) if @styles.include?(id)
  end
  inherited[:style].merge!(import_style(data[:Style])) if data[:Style]
  # AbstractView or Geometry: Use walk_tree to look up the method.
  entities = walk_tree(data, inherited)
  if data[:TimeStamp] || data[:TimeSpan]
    entities.flatten.each{ |entity|
      import_time(data, entity) if entity.is_a?(Sketchup::Entity)
    }
  end
  return entities
end



def import_model(data, inherited={})
  # Collect all data.
  name = inherited[:name]
  description = inherited[:description]
  lon = data[:Location][:longitude].to_f
  lat = data[:Location][:latitude].to_f
  alt = data[:Location][:altitude].to_f
  mode = data[:altitudeMode].to_sym if data[:altitudeMode]
  position = @model.latlong_to_point([lon, lat, alt])
  position.z = get_height(position, mode)
  if data[:Orientation]
    heading = data[:Orientation][:heading].to_f
    tilt = data[:Orientation][:tilt].to_f
    roll = data[:Orientation][:roll].to_f
  else
    heading, tilt, roll = 0
  end
  if data[:Scale]
    x = (data[:Scale][:x] || 1).to_f # Make sure that scale won't be 0.
    y = (data[:Scale][:y] || 1).to_f
    z = (data[:Scale][:z] || 1).to_f
  else
    x, y, z = 1
  end
  href = data[:Link][:href].to_s
  # Import the Collada file.
  raise(KMLImportError, "Reference to resource missing.", data.inspect) if href.empty?
  raise(KMLImportError, "Resource is not a local file. Only local files can be imported.", href) if href[/^http|^https|^ftp/]
  path = File.expand_path(File.join(@file_dir, href.sub(/^\.\//,"")))
  raise(KMLImportError, "File does not exist or is not a Collada model.", path) unless File.exists?(path) && File.extname(path) == ".dae"
  cache = @model.entities.to_a
  success = @model.import(path, false)
  raise(KMLImportError, "No model was imported.", path) unless success
  puts cache.inspect
  puts((@model.entities.to_a - cache).inspect)
  imported_instance = (@model.entities.to_a - cache).first
  raise(KMLImportError, "Imported model was not found.", path) if imported_instance.nil?
  # Apply data.
  h = Geom::Transformation.new([0,0,0], [0,0,1], heading.degrees)
  t = Geom::Transformation.new([0,0,0], [1,0,0], tilt.degrees)
  r = Geom::Transformation.new([0,0,0], [0,1,0], roll.degrees)
  l = Geom::Transformation.translation(position)
  s = Geom::Transformation.scaling(x,y,z)
  imported_instance.transform!( l * r * t * h * s )
  imported_instance.name = name
  imported_instance.definition.description = description if description
  imported_instance.definition.name = name
  set_attributes(imported_instance, data, "Model")
  return imported_instance
end # def import_model



def import_polygon(data, inherited={})
  # Collect all data.
  name = inherited[:name]
  alt_offset = data[:"gx:altitudeOffset"].to_f
  extrude = data[:extrude].to_i
  extrusion = 0
  mode = data[:altitudeMode].to_sym if data[:altitudeMode]

  # Parse outer loops.
  outer = data[:outerBoundaryIs][:LinearRing][:coordinates].to_s
  outer_loop = outer.scan(/([e0-9\.-]+),([e0-9\.-]+),?([e0-9\.-]+)?/m).
  map{ |p|
    # Convert strings into floats.
    p.map!{ |c| c.to_f }
    # Convert global into model coordinates.
    p = @model.latlong_to_point(p)
    # Set z according to altitudeMode.
    p.z = get_height(p, mode) + alt_offset.m
    p
  }.compact
  raise(KMLImportError, "Polygon contained less than 3 points.", data.inspect) if outer_loop.length < 3
  # Average terrain altitude + extrusion; sum up all z values and later divide by amount of summands.
  average_z = outer_loop.inject(0){ |sum, p| sum + p.z } / outer_loop.length.to_f
  # Lowest terrain altitude, necessary for extruded polygons in order to touch the ground.
  min_ter = outer_loop.map{ |p| get_terrain_elevation(p) }.min # min{ |z1, z2| z1 <=> z2 }

  # Make sure the polygon is planar. SketchUp creates only planar faces and we
  # don't want to tesselate the face over the terrain.
  plane = Utilities.points_to_plane(outer_loop)
  if extrude == 1 && inherited[:style] && inherited[:style][:facefill] != 0
    outer_loop.each{ |p| p.z = average_z }
  else
    outer_loop.map!{ |p| p.project_to_plane(plane) }
  end

  # Parse inner loops.
  inner = data[:innerBoundaryIs]
  unless inner.nil? || inner.empty?
    inner_loops = inner[:LinearRing]
    inner_loops = [inner_loops] unless inner_loops.is_a?(Array)
    inner_loops.map!{ |l|
      l = l[:coordinates].scan(/([e0-9\.-]+),([e0-9\.-]+),?([e0-9\.-]+)?/m).
      map{ |p|
        # Convert strings into floats.
        p.map!{ |c| c.to_f }
        # Convert global into model coordinates.
        p = @model.latlong_to_point(p)
      }.compact
      # Set z according to outer_loop.
      if extrude == 1 && inherited[:style] && inherited[:style][:facefill] != 0
        l.each{ |p| p.z = average_z }
      else
        l.map!{ |p| p.project_to_plane(plane) }
      end
      l
    }
  end

  context = inherited[:current_context] || @model
  if @@options[:geometry_grouped]
    context = context.entities.add_group
    set_attributes(context, data, "Polygon")
  end
  cache = context.entities.to_a
  context.name ||= name

  # Draw outer loop.
  # Create edges only.
  if inherited[:style] && inherited[:style][:facefill] == 0
    # Copy the first coordinate to the end to close the loop.
    outer_loop.push(outer_loop.first)
    context.entities.add_edges(outer_loop)
  # Create face with edges.
  else
    # Delete consecutive duplicates.
    Utilities.delete_duplicates(outer_loop)
    begin
      face = context.entities.add_face(outer_loop)
      face.reverse! if face.normal.z == -1
    rescue Exception => e
      log("A face could not be created.", e.message, data.inspect)
    end
  end

  # Draw inner loops.
  if inner_loops
    # Create edges only.
    if inherited[:style] && inherited[:style][:facefill] == 0
      inner_loops.each{ |l|
        # Copy the first coordinate to the end to close the loop.
        l.push(l.first)
        context.entities.add_edges(l)
      }
    # Create face with edges.
    else
      inner_loops.each{ |l|
        next log("An 'innerBoundaryIs' contained less than 3 points and was ignored.", l.inspect, data.inspect) if l.length < 3
        begin
          Utilities.delete_duplicates(l)
          f = context.entities.add_face(l)
          context.entities.erase_entities(f)
        rescue Exception => e
          log("A hole could not be created.", e.message, data.inspect)
        end
      }
    end
  end

  # Extrude from polygon to lowest terrain altitude.
  if extrude == 1 && inherited[:style] && inherited[:style][:facefill] != 0
    extrusion = (average_z - min_ter).abs
    face.pushpull(-extrusion)
  end

  # Array that contains all created entities.
  created_entities = context.entities.to_a - cache

  # Colorize.
  if inherited[:style]
    if @@options[:geometry_grouped]
      context.material = inherited[:style][:facecolor]
    else
      created_entities.each{ |entity|
        if entity.is_a?(Sketchup::Edge)
          entity.material = inherited[:style][:linecolor]
        elsif entity.is_a?(Sketchup::Face)
          entity.material = inherited[:style][:facecolor]
          entity.back_material = inherited[:style][:facecolor]
        end
      }
    end
    @model.rendering_options["EdgeColorMode"] = 0 if inherited[:style][:linecolor]
  end

  return created_entities
end # def import_polygon



def import_line_string(data, inherited={})
  # Collect all data.
  name = inherited[:name]
  alt_offset = data[:"gx:altitudeOffset"].to_f
  mode = data[:altitudeMode].to_sym if data[:altitudeMode]
  # mode = :clampToGround if data[:tessellate].to_i == 1 # Tessellate has only an effect with clampToGround.
  extrude = data[:extrude].to_i

  # Parse coordinates.
  coordinates = data[:coordinates].to_s
  ps = coordinates.scan(/([e0-9\.-]+),([e0-9\.-]+),?([e0-9\.-]+)?/m).
  map{ |p|
    # Convert strings into floats.
    p.map!{ |c| c.to_f }
    # Convert global into model coordinates.
    p = @model.latlong_to_point(p)
    # Set z according to altitudeMode.
    p.z = get_height(p, mode) + alt_offset.m
    p
  }
  raise(KMLImportError, "Linestring contained less than 2 points.", data.inspect) if ps.length < 2

  context = inherited[:current_context] || @model
  if @@options[:geometry_grouped]
    context = context.entities.add_group
    set_attributes(context, data, "LineString")
  end
  context.name ||= name

  # Draw curve from points.
  created_entities = context.entities.add_curve(ps)
  # Draw extrusion
  if extrude == 1
    pts = ps.map{ |p|
      p = p.clone
      p.z = get_terrain_elevation(p)
      p
    }
    unless pts == ps
      context.entities.add_edges(ps.first, pts.first)
      (1...ps.length).each{ |i|
        created_entities << face = context.entities.add_face(ps[i], pts[i], pts[i-1], ps[i-1])
        # Smooth inbetween edges
        unless i == ps.length - 1
          face.edges.first.smooth = true
          face.edges.first.soft = true
        end
      }
    end
  end

  # Colorize.
  if inherited[:style]
    if @@options[:geometry_grouped] && extrude != 1
      context.material = inherited[:style][:linecolor]
    else
      created_entities.each{ |entity|
        if entity.is_a?(Sketchup::Edge)
          entity.material = inherited[:style][:linecolor]
        elsif entity.is_a?(Sketchup::Face)
          entity.material = inherited[:style][:facecolor]
          entity.back_material = inherited[:style][:facecolor]
        end
      }
    end
    @model.rendering_options["EdgeColorMode"] = 0 if inherited[:style][:linecolor]
  end

  return created_entities
end # def import_line_string



def import_track(data, inherited={})
  # Collect all data.
  name = inherited[:name]
  mode = data[:altitudeMode].to_sym if data[:altitudeMode]
  model = data[:Model]

  # Parse coordinates.
  # <gx:coord> should be in all valid cases an array. There should not exist a track with only one point.
  ps = data[:"gx:coord"].map{ |p|
    p = p.scan(/[e0-9\.-]+/)
    # Convert strings into floats.
    p.map!{ |c| c.to_f }
    # Convert global into model coordinates.
    p = @model.latlong_to_point(p)
    # Set z according to altitudeMode.
    p.z = get_height(p, mode)
    p
  }

  context = inherited[:current_context] || @model
  if @@options[:geometry_grouped]
    context = context.entities.add_group
    set_attributes(context, data, "Track")
  end
  context.name ||= name

  # Draw curve from points.
  created_entities = context.entities.add_curve(ps)

  # Colorize.
  if inherited[:style]
    if @@options[:geometry_grouped]
      context.material = inherited[:style][:linecolor]
    else
      created_entities.each{ |edge| edge.material = inherited[:style][:linecolor] }
    end
    @model.rendering_options["EdgeColorMode"] = 0 if  inherited[:style][:linecolor]
  end

  # TODO: time and angle attributes.

  # Import a model if available.
  if model
    imported = import_model(model, inherited)
    p = Geom::Transformation.translation(ps.first)
    imported.transform!(p)
    created_entities << imported
  end

  return created_entities
end # def import_track



def import_point(data, inherited={})
  # Collect all data.
  name = inherited[:name]
  description = inherited[:description].to_s
  mode = data[:altitudeMode].to_sym if data[:altitudeMode]

  # Parse coordinates.
  coordinates = data[:coordinates].to_s
  p = coordinates.scan(/([e0-9\.-]+),([e0-9\.-]+),?([e0-9\.-]+)?/m).first.
  # Convert strings into floats.
  map{ |c| c.to_f }
  # Convert global into model coordinates.
  p = @model.latlong_to_point(p)
  # Set z according to altitudeMode.
  p.z = get_height(p, mode)

  context = inherited[:current_context] || @model
  created_entities = []

  # Draw points and texts from points.
  if description.is_a?(String)
    # Limit the length of added texts.
    # TODO: Better if we can set the text size and visibility depending on the viewport.
    description = description.slice(/^[^\n\r]{1,50}/).to_s
    created_entities << context.entities.add_text(description, p, [1,1,1])
  elsif name && !name.empty?
    created_entities << context.entities.add_text(name, p, [1,1,1])
  end
  point = context.entities.add_cpoint(p)
  created_entities << point

  set_attributes(point, data, "Point")

  return created_entities
end # def import_point



def import_ground_overlay(data, inherited={})
  # Collect all data.
  name = data[:name].to_s
  description = data[:description]
  href = data[:Icon][:href]
  raise(KMLImportError, "Reference to resource missing.", data.inspect) if href.empty?
  raise(KMLImportError, "Resource is not a local file. Only local files can be imported.", href) if href[/^http|^https|^ftp/]
  path = File.expand_path(File.join(@file_dir, href.sub(/^\.\//,"")))
  raise(KMLImportError, "File does not exist.", path) unless File.exists?(path)

  context = inherited[:current_context] || @model

  # Case 1: LatLonQuad with 4 corner points.
  if data[:"gx:LatLonQuad"]
    coordinates = data[:"gx:LatLonQuad"][:coordinates].to_s
    ps = coordinates.scan(/([e0-9\.-]+),([e0-9\.-]+)/).
    map{ |c|
      # Convert global into model coordinates.
      c = @model.latlong_to_point( [c[0].to_f, c[1].to_f, 0] )
      # Set z according to 0.
      c.z = 0
      c
    }
    # Put the last element at the begin.
    ps.unshift(ps.pop)

    # Draw grouped quadface.
    image = context.entities.add_group
    set_attributes(image, data, "GroundOverlay")
    image.name = name
    image.description = description if description
    face = image.entities.add_face(*ps)
    face.reverse! if face.normal.z < 0
    material = @model.materials.add(name)
    material.texture = path
    face.material = material
    pt_array = [ps[0], [0,1,0], ps[1], [0,0,0], ps[2], [1,0,0], ps[3], [1,1,0]]
    face.position_material(material, pt_array, true)

  # Case 2: N, S, E, W, rotation.
  else
    north = data[:LatLonBox][:north].to_f
    south = data[:LatLonBox][:south].to_f
    east = data[:LatLonBox][:east].to_f
    west = data[:LatLonBox][:west].to_f
    rotation = data[:LatLonBox][:rotation].to_f
    # Calculate width/height from 3 points.
    p1 = @model.latlong_to_point([west, south])
    p2 = @model.latlong_to_point([west, north])
    p3 = @model.latlong_to_point([east, north])
    center = @model.latlong_to_point([(west + east)/2, (north + south)/2, 0])
    w = p2.distance(p3)
    h = p2.distance(p1)

    # Draw a rectangular image.
    image = context.entities.add_image(path, p1, w, h)
    set_attributes(image, data, "GroundOverlay")
    raise(KMLImportError, "Creating Sketchup::Image failed.", [path, p1, w, h]) unless image.is_a?(Sketchup::Image)
    rotation_correction = -Math.acos((p2.y - p1.y)/h)
    tr_rotation_correction = Geom::Transformation.rotation(p1, [0,0,1], rotation_correction)
    image.transform!(tr_rotation_correction)
    tr_rotation = Geom::Transformation.rotation(center, [0,0,1], rotation.degrees)
    image.transform!(tr_rotation)
    Utilities::Image.get_definition(image).description = description if description
  end

  return image
end # def import_ground_overlay



def import_photo_overlay(data, inherited={})
  # Collect all data.
  name = data[:name].to_s
  description = data[:description]
  href = data[:Icon][:href]
  raise(KMLImportError, "Reference to resource missing.", data.inspect) if href.empty?
  raise(KMLImportError, "Resource is not a local file. Only local files can be imported.", href) if href[/^http|^https|^ftp/]
  path = File.expand_path(File.join(@file_dir, href.sub(/^\.\//,"")))
  raise(KMLImportError, "File does not exist.", path) unless File.exists?(path)
  left_fov = data[:ViewVolume][:leftFov].to_f
  right_fov = data[:ViewVolume][:rightFov].to_f
  bottom_fov = data[:ViewVolume][:bottomFov].to_f
  top_fov = data[:ViewVolume][:topFov].to_f

  # Create camera.
  page = import_camera(data[:Camera]) if data[:Camera]
  page = import_look_at(data[:LookAt]) if data[:LookAt]
  if page
    camera = page.camera
    @model.pages.erase(page)
  else
    camera = Sketchup::Camera.new
  end
  camera.fov = (top_fov - bottom_fov + right_fov - left_fov) / 2
  # camera.aspect_ratio = (left_fov + right_fov) / (bottom_fov + top_fov)

  # Create match photo scene.
  page = @model.pages.add_matchphoto_page(path, camera, name)
  page.description = description if description
  set_attributes(page, data, "PhotoOverlay")
  return page
end # def import_photo_overlay



def import_camera(data, inherited={})
  # Collect all data.
  mode = data[:altitudeMode].to_sym if data[:altitudeMode]
  lon = data[:longitude].to_f
  lat = data[:latitude].to_f
  alt = data[:altitude].to_f
  heading = data[:heading].to_f
  tilt = data[:tilt].to_f
  roll = data[:roll].to_f

  # Calculate camera parameters.
  eye = @model.latlong_to_point([lon, lat, alt])
  eye.z = get_height(eye, mode)
  x = Math.sin(heading.degrees)
  y = Math.cos(heading.degrees)
  z = Math.sin(-0.5*Math::PI + tilt.degrees)
  vector = Geom::Vector3d.new(x, y, z)
  target = eye + vector
  trans_roll = Geom::Transformation.new(eye, [x,y,z], roll.degrees)
  up = Z_AXIS.transform(trans_roll)

  # Create camera.
  camera = Sketchup::Camera.new(eye, target, up)
  page = @model.pages.add(inherited[:name])
  page.description = inherited[:description] if inherited[:description]
  page.camera.set(camera.eye, camera.target, camera.up)
  set_attributes(page, data, "Camera")
  return page
end # def import_camera



def import_look_at(data, inherited={})
  # Collect all data.
  mode = data[:altitudeMode].to_sym if data[:altitudeMode]
  lon = data[:longitude].to_f
  lat = data[:latitude].to_f
  alt = data[:altitude].to_f
  heading = data[:heading].to_f
  tilt = data[:tilt].to_f
  range = data[:range].to_f

  # Calculate camera parameters.
  target = @model.latlong_to_point([lon, lat, alt])
  target.z = get_height(target, mode)
  x = Math.sin(heading.degrees)
  y = Math.cos(heading.degrees)
  z = Math.sin(-0.5*Math::PI + tilt.degrees)
  vector = Geom::Vector3d.new(x, y, z)
  vector.length = range.m.to_inch
  eye = target - vector
  up = Z_AXIS

  # Create camera.
  camera = Sketchup::Camera.new(eye, target, up)
  page = @model.pages.add(inherited[:name])
  page.description = inherited[:description] if inherited[:description]
  page.camera.set(camera.eye, camera.target, camera.up)
  set_attributes(page, data, "LookAt")
  return page
end # def import_look_at



def import_tour(data, inherited={})
  name = data[:name].to_s
  description = data[:description]

  fly_tos = data[:"gx:FlyTo"]
  pages = []
  page_counter = 0
  fly_tos.each{ |fly_to|
    transition = fly_to[:"gx:duration"]
    inherited[:name] = "#{name} #{page_counter+=1}"
    cam = fly_to[:Camera]
    page = import_camera(cam, inherited) unless cam.empty?
    look_at = fly_to[:LookAt]
    page = import_look_at(look, inherited) unless look_at.empty?
    # is_wait = !fly_to[:"gx:Wait"].nil? # TODO
    page.transition_time = transition.to_f if page && transition
    pages << page
  }
  return pages
end # import_tour



#########################        helper methods        #########################



# Georeference the model.
# @param [Float] lon Longitude
# @param [Float] lat Latitude
# @param [Float] alt Altitude
#
def set_geo_reference(lon, lat, alt=0)
  @model.set_attribute("GeoReference", "Longitude", lon)
  @model.set_attribute("GeoReference", "Latitude", lat)
  @model.set_attribute("GeoReference", "Altitude", alt)
  @model.set_attribute("GeoReference", "LocationSource", "Manual")
  @model.set_attribute("GeoReference", "ZValueCentered", -alt.m)
  @model.set_attribute("GeoReference", "ModelTranslationZ", -alt.m)
  @model.set_attribute("GeoReference", "ModelTranslationY", 0) # TODO: SketchUp 7 requires a value here (?), but what does the value mean?
  @model.set_attribute("GeoReference", "ModelTranslationX", 0) # TODO
  @model.set_attribute("GeoReference", "UsesGeoReferencing", true)
  shadow_info = @model.shadow_info
  shadow_info["City"] = "Unknown"
  shadow_info["Country"] = "Unknown"
  shadow_info["Longitude"] = lon
  shadow_info["Latitude"] = lat
end # def set_geo_reference



# Calculate altitude of a point, considering altitudeMode.
# (everything in model coordinates)
# @param [Geom::Point3d,Array<Numeric>] point in the model, z does not follow mode
# @param [Symbol] mode
# @returns [Float] the altitude in meters according to mode
#
def get_height(point, mode)
  alt = point.z.to_f
  if mode == :clampToGround || mode.nil?
    ter = get_terrain_elevation(point)
    return ter
  elsif mode == :relativeToGround
    ter = get_terrain_elevation(point)
    return ter + alt
  elsif mode == :absolute
    return alt - @elevation.m
  else # TODO: relativeToSeaFloor, clampToSeaFloor (requires gx:altitudeMode)
    return alt - @elevation.m
  end
end # def get_height



# Get the elevation of the Google Earth Terrain (in model coordinates)
# @param [Geom::Point3d,Array<Numeric>] point
# @returns [Float] the elevation at the given point
#
def get_terrain_elevation(point)
  p = Geom::Point3d.new(point)
  p.z = @model.bounds.min.z
  # Raytest:
  # Do not use [ray]=[point,vector] but point,vector,wysiwyg_flag
  # because SU<8M1 interprets two parameters also as a ray
  raytest = @model.raytest(p, Z_AXIS, false)
  until raytest.nil? || raytest[1][0].is_a?(Sketchup::Group) && raytest[1][0].name[/Google Earth Terrain/]
      p.z = raytest[0].z
      raytest = @model.raytest(p, Z_AXIS, false)
  end
  return (!raytest.nil?) ? raytest[0].z : 0
end # def get_terrain_elevation



# Converts Google Earth's hexadecimal aabbggrr colors to SketchUp's rgba.
# @param [String] kml_color
# @returns [Sketchup::Material]
#
def color_to_su(kml_color) # accepts material or color
  return nil unless kml_color.is_a?(String)
  rgba = kml_color.scan(/[a-fA-F0-9]{2}/).map{ |x| x.hex }.reverse!
  rgba.push(255) while rgba.length < 4
  alpha = (100 * rgba.pop / 255.to_f).round / 100.to_f
  color = Sketchup::Color.new(rgba)
  # Try to find an existing material.
  material = @model.materials[kml_color]
  if material.nil?
    material = @model.materials.find{ |m|
      m.color == color && m.alpha == alpha
    }
  end
  # Otherwise create a new material.
  if material.nil?
    material = @model.materials.add(kml_color)
    material.color = color
    material.alpha = alpha
  end
  return material
end # def color_to_su



def log(*message)
  @log << message
end



def raise(error, reason, debug_data)
  file_line = caller.first.to_s[/[^\/\:]+\:\d+(?=\:)/]
  Kernel.raise(error, "#{reason} (#{file_line})", debug_data)
end



class KMLImportError < StandardError
end



end # class Importer



module Utilities



# Calculates the average plane between more or less precise positioned points (or random points)
# @param [Array<Geom::Point3d>] array
# @returns [Array<Geom::Point3d,Geom::Vector3d] plane
#
def self.points_to_plane(array)
  a = Array.new(array.compact)
  while a.length > 3
    b = a.shift
    c = a.shift
    d = []
    (0..2).each{|i| d[i] = (b[i]+c[i])/2.to_f }
    a.push(d)
  end
  vector1 = a[0].vector_to(a[1])
  vector2 = a[1].vector_to(a[2])
  # vector2 = vector2.to_a.reverse if vector1.parallel? vector2
  plane = [a[0], vector1*vector2]
  return plane
end



# Check if an array of points forms a plane or not.
# @param [Array<Geom::Point3d>] array
# @returns [Boolean]
#
def self.plane?(array)
  plane = [Geom::Point3d.new(array[0]),array[1].vector_to(array[2])]
  (3..array.length-1).each{|i|
    return false if (array[i]-a[0]).on_plane?(plane) == false
  }
  return true
end



# Remove consecutive duplicates from an array.
# @param [Array] array
# @returns [Array]
#
def self.delete_duplicates(array)
  to_delete = []
  (0..array.length).each{ |i|
    to_delete << i if array[i] == array[i-1]
  }
  to_delete.reverse.each{ |i|
    array.delete_at(i)
  }
  return array
end



# Splits an array a after every position of a duplicate array element
# and returns an array of subarrays that all don't have duplicates anymore.
# @param [Array] a
# @returns [Array<Array>]
#
def self.split_uniq(a)
  b = []
  i = 0
  j = 1
  until i == a.length
    until j == a.length
      if a[i] == a[j]
        c = a.slice!(0..(j-1))
        b.push(c)
        j = 1
      else
        j += 1
      end
    end
    i += 1
    j = i + 1
  end
  b.push(a) if a.length != 0
  return b
end # def split_uniq



end # module Utilities



unless file_loaded?(__FILE__)
  Sketchup.register_importer(Importer.new)

  # Command for KML Import
  cmd_kmlimporter = UI::Command.new(TRANSLATE["Import KML File"]) {
    begin
      importer = Importer.new
      dialog = importer.do_options
      block = Proc.new{
        path = UI.openpanel(TRANSLATE["Open KML file"], ".kml")
        next unless path # If file dialog has been closed/cancelled.
        success = importer.load_file(path)
        case success
        when 1
          UI.messagebox(TRANSLATE["File seems not to exist.\n Please check that the file path does not contain Unicode characters. This is a limitation of SketchUp and its Ruby version."], MB_OK)
        when 4
          UI.messagebox(TRANSLATE["Import failed."], MB_OK)
        end
      }
      # Attach an event handler to when the dialog is closed.
      if dialog.visible?
        dialog.on_close(&block)
      # If it is a modal dialog (SketchUp on Windows), it has halted execution
      # until it was closed, so we can directly run the exporter.
      else
        block.call
      end
    rescue Exception => e
      $stderr.write(e.message)
      $stderr.write(e.backtrace.join("\n"))
    end
  }
  cmd_kmlimporter.tooltip = TRANSLATE["Imports geo-referenced data (placemarks, tracks, ground overlays etc.) from KML files into Sketchup."]
  cmd_kmlimporter.small_icon = File.join(DIR, "images", "icon_kmlimport_16.png")
  cmd_kmlimporter.large_icon = File.join(DIR, "images", "icon_kmlimport_24.png")
  self::MENU = UI.menu("Plugins").add_submenu("KML Tools") unless defined?(self::MENU)
  self::MENU.add_item(cmd_kmlimporter)

  file_loaded(__FILE__)
end



end # module KMLTools



end # module AE

