/* module KMLExporter:
 * Methods for the exporter of the SketchUp plugin KMLTools.
 */
AE.KMLExporter = function(self) {



var $ = AE.$;



// Whether the exporter is ready to start right now.
var ready = true;



self.initialize = function(options, features) {
  if (!options) { options = {}; }

  // Hide all features that are not available for export.
  // features = ["model", "curve", "hface", "himage", "photomatch", "cpoint", "text", "page"]
  var featureSections = $("#features").childNodes;
  for (var j=0; j<featureSections.length; j++) {
    if (featureSections[j].nodeType !== 1) { continue; }
    var o = featureSections[j];
    if (inArray(o.id, features)) {
      AE.Style.show(o);
      o.getElementsByTagName("input")[0].checked = true;
    } else {
      AE.Style.hide(o);
      o.getElementsByTagName("input")[0].checked = false;
    }
    o.getElementsByTagName("input")[0].onclick = function() { toggleDisabled(o); };
  };

  // Show message if the model contains no features for export.
  if (!features || features.length <= 1) {
    // Hide the contents of the form.
    var fieldsets = $("fieldset");
    for (var i=0; i<fieldsets.length; i++) {
      AE.Style.hide(fieldsets[i]);
    }
    // Show a message.
    AE.Style.show($("#info"));
    $("#info").innerHTML = AE.Translate.get("The model contains nothing to export.");
    $("#button_ok").onclick = function() {
      AE.Dialog.close();
    };
  }

  /* Load defaults.
   * Identifies input elements when their name matches the default option's name.
   */
  AE.Form.fill(options);

  // Loop over all nested inputs and apply an event hander for toggleDisabled.
  var inputs = $("input", $("#features"));
  for (var j=0; j<inputs.length; j++) {
    var input = inputs[j];
    if (input.type == "radio" || input.type == "checkbox") {
      input.onclick = function() { toggleDisabled(this); };
      toggleDisabled(input);
    };
  };

  // Ok button.
  $("#button_ok").onclick = function() {
    // Wait until ready.
    new AE.Scheduler(1).wait( function() { return ready; }, function() {
      AE.Bridge.updateOptions(AE.Form.read());
      AE.Dialog.close();
    });
  };

  // Prevent page reloading on buttons in form element.
  $("form")[0].onsubmit = function() { return false };

  // Adjust the size (height) when the dialog has been manually resized.
  // Internet Explorer it requires a little delay to complete resizing because it triggers
  // onresize twice (for height & width). Otherwise we would use incorrect
  // dimensions and the dialog flickers.
  window.onresize = function() { window.setTimeout(AE.Dialog.adjustSize, 0) };

  AE.Dialog.adjustSize();
};



/** Disable or enable input elements in cascaded forms.
 *    Args:
 *      input:
 *      1) A checkbox or radio button that has an rel="c1" attribute pointing at
 *         the id of another input/container that should be disabled/enabled.
 *         <somewhere> <checkbox rel="c1"> </somewhere> ...
 *         <somewhereelse> <container id="c1"> <input> </container> </somewhereelse>
 *      2) A container that has only one checkbox/radio button as direct child element.
 *         All other inputs cascaded in subcontainers of this container will be disabled/enabled.
 *         <container> <checkbox> <subcontainer1> <input> </subcontainer1> </container>
 *     Returns:
 *       nothing
 */
var toggleDisabled = function(input) {
  //
  var toggleChildren = function(input) {
    // Look for rel attribute.
    if (input.getAttribute('rel')) {
      // Get the container.
      var relContainer = $("#" + input.getAttribute('rel'));
      relContainer.disabled = !input.checked;
      // Loop over all children of relContainer and set their disabled status.
      var relInputs = $('input', relContainer);
      for (var i=0; i<relInputs.length; i++) {
        relInputs[i].disabled = !input.checked;
      }
    // Parent element must have only one child input and several sub-containers.
    } else {
      // Get all containers that should be enabled/disabled.
      var c = input.parentNode.childNodes
      // If this input is wrapped in a label, then the parent that we want is outside.
      if (input.parentNode.nodeName == "LABEL") {
        c = input.parentNode.parentNode.childNodes;
      }
      var n = []; // Collects a boolean input (Should only contain one element.)
      // Get the boolean input that decides over the subcontainers' disabled state.
      for (var i=0; i<c.length; i++) {
        // Boolean input (checkbox/radio).
        if (c[i].nodeName == "INPUT" &&
            (c[i].type == "checkbox" || c[i].type=="radio")) {
          n.push(c[i]);
        }
        // Nested boolean input inside a label (checkbox/radio).
        else if (c[i].nodeName == "LABEL") {
          var d = c[i].childNodes;
          for (var j=0; j<d.length; j++) {
            if (d[j].nodeName == "INPUT" &&
                (d[j].type == "checkbox" || d[j].type == "radio")) {
              n.push(d[j]);
              break;
            }
          }
        }
      }
      // Toggle the disabled state for all inputs in all containers according to n[0].
      if (n.length == 1) {
        for (var i=0; i<c.length; i++) {
          if (c[i].childNodes.length == 0) { continue; }
          var ci = c[i].getElementsByTagName('input');
          for (var k=0; k<ci.length; k++) {
            if (!(ci[k] == input)) {
              ci[k].disabled = !input.checked;
            }
          }
        }
      }
    }
  }
  toggleChildren(input);
  // radio: Loop over all connected radio buttons (with same name).
  if (input.type == 'radio') {
    var radios = document.getElementsByName(input.name);
    for (var i=0; i<radios.length; i++) {
      if (radios[i] != input) { toggleChildren(radios[i]); }
    }
  }
};



// Function to get elevation data from a web service in case it is not contained in the model.
self.getElevation = function(lat, lon) {
  // Lock the exporter to wait until the elevation is returned.
  ready = false;
  // Fallback if json request is unsuccessful.
  window.setTimeout(function(){ ready = true; }, 5000); // TODO: No need to wait if Sketchup.is_online? == false
  // Request json data from geonames.org astergdem API.
  var callback = "AE.KMLExporter.receiveElevation";
  var url = "http://api.geonames.org/astergdemJSON?formatted=true&lat="
  url += lat + "&lng=" + lon + "&username=aerilius&style=full&callback=" + callback;
  getJson(url);
};



self.receiveElevation = function(json) {
  AE.Bridge.callRuby("receive_elevation", json["astergdem"]);
  ready = true;
};



var getJson = function(url) {
  var s = document.createElement("script");
  s.setAttribute("src", url);
  document.body.appendChild(s);
  void(0);
};



/* module Progress
 * A primitive progress indicator.
 */
/*
self.Progress = function(progress) {

  var length = 1;
  var state = 0;

  // Show progress animation.
  progress.start = function(n) {
    if (n != null && n != 0) {
      length = n;
      state = 0;
    }
    AE.Style.show( $("#overlay") );
    $("progress").style.width = 0;
  }

  progress.advance = function(steps, processname) {
    if (processname) {
      $("#progressbar-action").innerHTML = processname;
    }
    state += (typeof steps == "number") ? steps : 1;
    var percent = Math.round( 100*state/length );
    if (percent > 100) { percent = 100; }
    $("progress").style.width = percent + "%";
  }

  // Hide progress animation.
  progress.end = function() {
    length = 1;
    state = 0;
    AE.Style.hide( $("#overlay") );
  }

  return progress;
}(self.Progress || {}); // end module Progress
*/



var inArray = function(value, array) {
  for (var i=0; i<array.length; i++) {
    if (array[i] == value) { return true; }
  };
  return false;
};



return self;
}(AE.KMLExporter || {}); // end module KMLExporter
